#include <windows.h>
#include <gl\glew.h>
#include <gl\gl.h>
#include <gl\glu.h>
#include <gl\freeglut.h>
#include <glm/glm.hpp>
#include <glm/gtc/matrix_transform.hpp>
#include <glm/gtx/transform2.hpp>
#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "file_reader.h"

using namespace std;


// kparny
float aspect_ratio = 1.0;

// billentyk llapotai
bool keys[256];

// vertex shader, fragment shader s a shader program azonosti
GLuint v,f,p;

// textra
TextureImage img;

// vertex array object kezelje a kirajzoland objektumhoz
GLuint vaoHandle;

// kamera paramterei
float cameraRotateX = 0;
float cameraRotateY = -45.0f;
glm::vec3 cameraTranslate(-10.0f, -0.0f, -10.0f);

// objektum paramterei
glm::vec3 objectRotate;


void changeSize(int w, int h) 
{
	if (h == 0)
	{
		h = 1;
	}

	aspect_ratio =  w * 1.0f / h;
	glViewport(0, 0, w, h);
}


// shader-ek belltsa
void setShaders() 
{
	char *vs = NULL, *fs = NULL;

	// shader-ek ltrehozsa
	v = glCreateShader(GL_VERTEX_SHADER);
	if( v == 0 )
	{
		fprintf(stderr, "Error creating vertex shader.\n");
		exit(1);
	}
	f = glCreateShader(GL_FRAGMENT_SHADER);
	if( f == 0 )
	{
		fprintf(stderr, "Error creating fragment shader.\n");
		exit(1);
	}

	// shader-ek forrskdjnak beolvassa
	vs = textFileRead("reflect.vert");
	fs = textFileRead("reflect.frag");

	const char * ff = fs;
	const char * vv = vs;

	// forrskd hozzcsatolsa a shader azonostkhoz
	glShaderSource(v, 1, &vv, NULL);
	glShaderSource(f, 1, &ff, NULL);

	free(vs);
	free(fs);

	// shader-ek fordtsa
	glCompileShader(v);
	glCompileShader(f);

	// vertex shader fordts sikeressgnek ellenrzse
	GLint result;
	glGetShaderiv(v, GL_COMPILE_STATUS, &result);
	if( GL_FALSE == result )
	{
		fprintf( stderr, "Vertex shader compilation failed!\n" );
		
		GLint logLen;
		glGetShaderiv(v, GL_INFO_LOG_LENGTH, &logLen);
		
		if( logLen > 0 )
		{
			char * log = (char *)malloc(logLen);
			
			GLsizei written;
			glGetShaderInfoLog(v, logLen, &written, log);
			fprintf(stderr, "Shader log:\n%s", log);
			free(log);
		}
	}

	// fragment shader fordts sikeressgnek ellenrzse
	result;
	glGetShaderiv(f, GL_COMPILE_STATUS, &result);
	if( GL_FALSE == result )
	{
		fprintf( stderr, "Fragment shader compilation failed!\n" );
		
		GLint logLen;
		glGetShaderiv(f, GL_INFO_LOG_LENGTH, &logLen);
		
		if( logLen > 0 )
		{
			char * log = (char *)malloc(logLen);
			
			GLsizei written;
			glGetShaderInfoLog(f, logLen, &written, log);
			fprintf(stderr, "Shader log:\n%s", log);
			free(log);
		}
	}

	// shader program ltrehozsa s a lefordtott shader-ek hozzcsatolsa
	p = glCreateProgram();
	if( p == 0 )
	{
		fprintf(stderr, "Error creating program object.\n");
		exit(1);
	}

	glAttachShader(p, f);
	glAttachShader(p, v);

	// shader-ek sszeszerkesztse egy programm
	glLinkProgram(p);

	// sszeszerkeszts sikeressgnek ellenrzse
	GLint status;
	glGetProgramiv( p, GL_LINK_STATUS, &status);
	if( GL_FALSE == status ) 
	{
		fprintf( stderr, "Failed to link shader program!\n" );
		GLint logLen;
		glGetProgramiv(p, GL_INFO_LOG_LENGTH, &logLen);
	
		if( logLen > 0 )
		{
			char * log = (char *)malloc(logLen);
			GLsizei written;
			glGetProgramInfoLog(p, logLen, &written, log);
			fprintf(stderr, "Program log: \n%s", log);
			free(log);
		}
	}
	else
	{
		// shader program hasznlatba vtele
		glUseProgram(p);
	}
}


// cube map textra betltse s ltrehozsa
void loadCubeMapTexture(char* baseFileName)
{
	glActiveTexture(GL_TEXTURE0);
	
	glGenTextures(1, &img.texID);
	glEnable(GL_TEXTURE_CUBE_MAP);
	glBindTexture(GL_TEXTURE_CUBE_MAP, img.texID);

	const char * suffixes[] = {"posx", "negx", "posy", "negy", "posz", "negz"};

	GLuint targets[] = { GL_TEXTURE_CUBE_MAP_POSITIVE_X,
						 GL_TEXTURE_CUBE_MAP_NEGATIVE_X,
						 GL_TEXTURE_CUBE_MAP_POSITIVE_Y,
						 GL_TEXTURE_CUBE_MAP_NEGATIVE_Y,
						 GL_TEXTURE_CUBE_MAP_POSITIVE_Z,
						 GL_TEXTURE_CUBE_MAP_NEGATIVE_Z };

	for (int i = 0; i < 6; i++) 
	{
		string texName = "images/";
		
		texName += string(baseFileName);
		texName += "_";
		texName += suffixes[i];
		texName += ".tga";

		if ( !loadTGA(&img, texName.c_str()) )
		{
			exit(1);
		}

		glTexImage2D(targets[i], 0, GL_RGB, img.width, img.height, 0, GL_RGB, GL_UNSIGNED_BYTE, img.imageData);
	}

	// cube map textra paramterek belltsa
	glTexParameterf(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameterf(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameterf(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameterf(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	glTexParameterf(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_EDGE);

	GLuint location = glGetUniformLocation(p, "CubeMapTex");
	if( location >= 0 )
	{
		glUniform1i(location, 0);
	}
}


void createObject()
{
	// pozci adatok
	float positionData[108] = {	 1, 1, 1,  -1, 1, 1,  -1,-1, 1,      // v0-v1-v2 (front)
								-1,-1, 1,   1,-1, 1,   1, 1, 1,      // v2-v3-v0
								 1, 1, 1,   1,-1, 1,   1,-1,-1,      // v0-v3-v4 (right)
								 1,-1,-1,   1, 1,-1,   1, 1, 1,      // v4-v5-v0
								 1, 1, 1,   1, 1,-1,  -1, 1,-1,      // v0-v5-v6 (top)
								-1, 1,-1,  -1, 1, 1,   1, 1, 1,      // v6-v1-v0
								-1, 1, 1,  -1, 1,-1,  -1,-1,-1,      // v1-v6-v7 (left)
								-1,-1,-1,  -1,-1, 1,  -1, 1, 1,      // v7-v2-v1
								-1,-1,-1,   1,-1,-1,   1,-1, 1,      // v7-v4-v3 (bottom)
								 1,-1, 1,  -1,-1, 1,  -1,-1,-1,      // v3-v2-v7
								 1,-1,-1,  -1,-1,-1,  -1, 1,-1,      // v4-v7-v6 (back)
								-1, 1,-1,   1, 1,-1,   1,-1,-1 };    // v6-v5-v4


	// norml vektor adatok
	float normalData[108]  = {  0, 0, 1,   0, 0, 1,   0, 0, 1,      // v0-v1-v2 (front)
								0, 0, 1,   0, 0, 1,   0, 0, 1,      // v2-v3-v0
								1, 0, 0,   1, 0, 0,   1, 0, 0,      // v0-v3-v4 (right)
								1, 0, 0,   1, 0, 0,   1, 0, 0,      // v4-v5-v0
								0, 1, 0,   0, 1, 0,   0, 1, 0,      // v0-v5-v6 (top)
								0, 1, 0,   0, 1, 0,   0, 1, 0,      // v6-v1-v0
							   -1, 0, 0,  -1, 0, 0,  -1, 0, 0,      // v1-v6-v7 (left)
							   -1, 0, 0,  -1, 0, 0,  -1, 0, 0,      // v7-v2-v1
								0,-1, 0,   0,-1, 0,   0,-1, 0,      // v7-v4-v3 (bottom)
								0,-1, 0,   0,-1, 0,   0,-1, 0,      // v3-v2-v7
								0, 0,-1,   0, 0,-1,   0, 0,-1,      // v4-v7-v6 (back)
								0, 0,-1,   0, 0,-1,   0, 0,-1 };    // v6-v5-v4

	// buffer object-ek ltrehozsa a pozcikhoz s normlvektorokhoz
	GLuint vboHandles[2];
	glGenBuffers(2, vboHandles);
	GLuint positionBufferHandle = vboHandles[0];
	GLuint normalBufferHandle = vboHandles[1];

	// a megfelel buffer object feltltse a pozci adatokkal 
	glBindBuffer(GL_ARRAY_BUFFER, positionBufferHandle);
	glBufferData(GL_ARRAY_BUFFER, 108 * sizeof(float), positionData, GL_STATIC_DRAW);

	// a megfelel buffer object feltltse a normlvektor adatokkal
	glBindBuffer(GL_ARRAY_BUFFER, normalBufferHandle);
	glBufferData(GL_ARRAY_BUFFER, 108 * sizeof(float), normalData, GL_STATIC_DRAW);

	// vertex array object ltrehozsa s belltsa
	glGenVertexArrays(1, &vaoHandle);
	glBindVertexArray(vaoHandle);

	// a vertex attribtumok tmbjeinek aktivlsa
	glEnableVertexAttribArray(0); 
	glEnableVertexAttribArray(1);

	// a 0-s index vertex attribtumhoz(azaz a pozcihoz) prostja a pozcikat trol buffer object-et
	glBindBuffer(GL_ARRAY_BUFFER, positionBufferHandle);
	glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 0, (GLubyte*) NULL);

	// az 1-es index vertex attribtumhoz(azaz a normlvektorhoz) prostja a normlvektorokat trol buffer object-et
	glBindBuffer(GL_ARRAY_BUFFER, normalBufferHandle);
	glVertexAttribPointer( 1, 3, GL_FLOAT, GL_FALSE, 0, (GLubyte*) NULL);
}


void initGL()
{
	glClearColor(1.0, 1.0, 1.0, 1.0);
	glEnable(GL_DEPTH_TEST);

	glewInit();
	if (glewIsSupported("GL_VERSION_4_0"))
	{
		printf("Ready for OpenGL 4.0\n");
	}
	else 
	{
		printf("OpenGL 4.0 not supported\n");
		exit(1);
	}

	setShaders();
	createObject();
	loadCubeMapTexture("landscape");

	for (int i = 0; i < 256; i++) { keys[i] = false; }

	// objektum sznnek belltsa
	GLuint location = glGetUniformLocation(p, "MaterialColor");
	if( location >= 0 )
	{
		glUniform4fv(location, 1, &glm::vec4(0.5, 0.5, 0.5, 1.0)[0]);
	}

	// tkrzds mrtknek belltsa
	location = glGetUniformLocation(p, "ReflectFactor");
	if( location >= 0 )
	{
		glUniform1f(location, 0.8);
	}
}


void renderScene(void) 
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	// projekcis mtrix
	glm::mat4 proj = glm::perspective(45.0f, aspect_ratio, 0.1f, 100.0f );

	// balra-jobbra nzs
	if ( keys['a'] )
	{
		cameraRotateY -= 3;		
	}
	else if ( keys['d'] )
	{
		cameraRotateY += 3;
	}
	
	// fel-le nzs
	if ( keys['w'] )
	{
		cameraRotateX += 3;		
	}
	else if ( keys['s'] )
	{
		cameraRotateX -= 3;
	}

	const float moveSpeed = 0.05;

	// mozgs az x tengelyen
	if ( keys['j'] )
	{
		cameraTranslate.x -= moveSpeed;		
	}
	else if ( keys['l'] )
	{
		cameraTranslate.x += moveSpeed;
	}

	// mozgs a z tengelyen
	if ( keys['i'] )
	{
		cameraTranslate.z += moveSpeed;		
	}
	else if ( keys['k'] )
	{
		cameraTranslate.z -= moveSpeed;
	}

	// mozgs az y tengelyen
	if ( keys['c'] )
	{
		cameraTranslate.y += moveSpeed;		
	}
	else if ( keys[' '] )
	{
		cameraTranslate.y -= moveSpeed;
	}

	// a skybox transzformcii
	glm::mat4 viewTranslate = glm::translate(glm::mat4(1.0f), glm::vec3(0.0f));
	glm::mat4 viewRotateX = glm::rotate(viewTranslate, cameraRotateX, glm::vec3(-1.0f, 0.0f, 0.0f));
	glm::mat4 view = glm::rotate(viewRotateX, cameraRotateY, glm::vec3(0.0f, 1.0f, 0.0f)); 

	glm::mat4 model(1.0f);
	model = glm::scale(model, glm::vec3(50, 50, 50));

	// a hrom mtrix szorzata adja a tnyleges transzformcit
	glm::mat4 mvp = proj * view * model;

	// a vertex shader MVP mtrixt belltja az elbb kiszmolt transzformcis mtrixra
	GLuint location = glGetUniformLocation(p, "MVP");
	if( location >= 0 )
	{
		glUniformMatrix4fv(location, 1, GL_FALSE, &mvp[0][0]);
	}

	// a vertex shader ModelMatrix mtrixt belltja a model mtrixra
	location = glGetUniformLocation(p, "ModelMatrix");
	if( location >= 0 )
	{
		glUniformMatrix4fv(location, 1, GL_FALSE, &model[0][0]);
	}

	// a vertex shader WorldCameraPosition vektorjt belltja a cameraTranslate vektorra
	location = glGetUniformLocation(p, "WorldCameraPosition");
	if( location >= 0 )
	{
		glUniform3fv(location, 1, &((-cameraTranslate)[0]));
	}

	// bellja a shader program DrawSkyBox vltozjt true-ra
	location = glGetUniformLocation(p, "DrawSkyBox");
	if( location >= 0 )
	{
		glUniform1i(location, 1);
	}

	// az skybox kirajzolsa
	glBindVertexArray(vaoHandle);
	glDrawArrays(GL_TRIANGLES, 0, 36);

	// ------------------------------------------------------------------------

	// az objektum traszformcii
	viewRotateX = glm::rotate(glm::mat4(1.0f), cameraRotateX, glm::vec3(-1.0f, 0.0f, 0.0f));
	view = glm::rotate(viewRotateX, cameraRotateY, glm::vec3(0.0f, 1.0f, 0.0f)); 
	view = glm::translate(view, cameraTranslate);

	// objectRotate.x += 0.1;
	objectRotate.y += 0.1;

	model = glm::mat4(1.0f);
	model = glm::rotate(model, objectRotate.x, glm::vec3(1.0, 0.0, 0.0));
	model = glm::rotate(model, objectRotate.y, glm::vec3(0.0, 1.0, 0.0));

	// a hrom mtrix szorzata adja a tnyleges transzformcit
	mvp = proj * view * model;

	// bellja a shader program DrawSkyBox vltozjt false-ra
	location = glGetUniformLocation(p, "DrawSkyBox");
	if( location >= 0 )
	{
		glUniform1i(location, 0);
	}

	// a vertex shader MVP mtrixt belltja az elbb kiszmolt transzformcis mtrixra
	location = glGetUniformLocation(p, "MVP");
	if( location >= 0 )
	{
		glUniformMatrix4fv(location, 1, GL_FALSE, &mvp[0][0]);
	}

	// a vertex shader ModelMatrix mtrixt belltja a model mtrixra
	location = glGetUniformLocation(p, "ModelMatrix");
	if( location >= 0 )
	{
		glUniformMatrix4fv(location, 1, GL_FALSE, &model[0][0]);
	}

	// tkrzd objektum kirajzolsa
	glBindVertexArray(vaoHandle);
	glDrawArrays(GL_TRIANGLES, 0, 36);

	glutSwapBuffers();
}


void normalKeyDown(unsigned char key, int x, int y) 
{
	keys[key] = true;
} 


void normalKeyUp(unsigned char key, int x, int y) 
{
	keys[key] = false;
}


void mouseMove(int x, int y) 
{
}


void mouseButton(int button, int state, int x, int y)
{
}


int main(int argc, char **argv) 
{
	// GLUT inicalizls s ablak ltrehozs
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_DEPTH | GLUT_DOUBLE | GLUT_RGBA);
	glutInitWindowPosition(0, 0);
	glutInitWindowSize(800, 600);
	glutCreateWindow("GLSL - Environment Mapping");

	// callback-ek belltsa
	glutDisplayFunc(renderScene);
	glutReshapeFunc(changeSize);
	glutIdleFunc(renderScene);

	glutKeyboardFunc(normalKeyDown);
	glutKeyboardUpFunc(normalKeyUp);
	glutMouseFunc(mouseButton);
	glutMotionFunc(mouseMove);
	glutPassiveMotionFunc(mouseMove);

	// OpenGL inicializls
	initGL();

	// glut fciklusba belps
	glutMainLoop();

	return 0;
}